-- StatRings Layout Code
---------------------------------------------------------------------------
-- The goal of this code is to come up with a framework for layout out
-- rings, where a combination of components are combined to instruct the
-- ring management code to create and configure the ring cluster:
--
-- LAYOUT MODEL
--   1) Each ring or component has a logical size, the layout manager will
--      work with each component to calculate an appropriate UI scale to 
--      give the desired scaling
--      (I think logical sizes should be fixed, at 200 x 200 / 100 radius)
--
--   2) Each component also has a logical scale, which is used when
--      determining the eventual UI scale.
--
--   3) I'm thinking of using a logo-esque layout language to describe
--      how to move from component to component, these movements will be
--      scripted with three scaling contexts also:
--          "O" - origin - the object from which the point is measured
--          "R" - root - the base object
--          "S" - self - the object being placed

local PARTY_ROTATION_STICKY = nil;
local PARTY_RING_GAP = 5;
local PARTY_SEP_ANGLE = 25;
local PARTY_OFS_ANGLE = 25/2;
local PET_OFFSET = -100;
local PET_SINK = -100;

StatRingsLayout_Specification = {
   root = "player", -- Specify which element is the root one

   objects = { -- List all of the configuration objects
      player = {
	 unit = "player",
	 
      },

      target = {
	 parent = "player", -- Parent object
	 scale = 0.9, -- Logical scale
      },

      pet = {
	 parent = "player",
	 scale = 0.4,
	 anchorPath = {
	    { 'move', 'O', PET_OFFSET }, -- Move ahead almost the whole radius of the ring
	    { 'move', 'S', -PET_SINK },
	 },
      },

      party1 = {
	 parent = "player",
	 scale = 0.25,
	 anchorPath = {
	    { 'turn', -(PARTY_OFS_ANGLE+PARTY_SEP_ANGLE),
	       PARTY_ROTATION_STICKY },
	    { 'move', 'O', 100 },
	    { 'move', 'R', PARTY_RING_GAP },
	    { 'move', 'S', 100 },
	 },
      },

      party2 = {
	 parent = "player",
	 scale = 0.25,
	 anchorPath = {
	    { 'turn', -PARTY_OFS_ANGLE, PARTY_ROTATION_STICKY },
	    { 'move', 'O', 100 },
	    { 'move', 'R', PARTY_RING_GAP },
	    { 'move', 'S', 100 },
	 },
      },

      party3 = {
	 parent = "player",
	 scale = 0.25,
	 anchorPath = {
	    { 'turn', PARTY_OFS_ANGLE, PARTY_ROTATION_STICKY },
	    { 'move', 'O', 100 },
	    { 'move', 'R', PARTY_RING_GAP },
	    { 'move', 'S', 100 },
	 },
      },

      party4 = {
	 parent = "player",
	 scale = 0.25,
	 anchorPath = {
	    { 'turn', (PARTY_OFS_ANGLE+PARTY_SEP_ANGLE)
	       , PARTY_ROTATION_STICKY },
	    { 'move', 'O', 100 },
	    { 'move', 'R', PARTY_RING_GAP },
	    { 'move', 'S', 100 },
	 },
      },

      partypet1 = {
	 parent = "party1",
	 scale = 0.4,
	 anchorPath = {
	    { 'move', 'O', PET_OFFSET }, -- Move ahead almost the whole radius of the ring
	    { 'move', 'S', -PET_SINK },
	 },
      },


      partypet2 = {
	 parent = "party2",
	 scale = 0.4,
	 anchorPath = {
	    { 'move', 'O', PET_OFFSET }, -- Move ahead almost the whole radius of the ring
	    { 'move', 'S', -PET_SINK },
	 },
      },

      partypet3 = {
	 parent = "party3",
	 scale = 0.4,
	 anchorPath = {
	    { 'move', 'O', PET_OFFSET }, -- Move ahead almost the whole radius of the ring
	    { 'move', 'S', -PET_SINK },
	 },
      },

      partypet4 = {
	 parent = "party4",
	 scale = 0.4,
	 anchorPath = {
	    { 'move', 'O', PET_OFFSET }, -- Move ahead almost the whole radius of the ring
	    { 'move', 'S', -PET_SINK },
	 },
      },
   },
}

local function DEBUG(msg) 
   if (io) then
      io.write("SRL: " .. msg .. "\n");
   else
      DEFAULT_CHAT_FRAME:AddMessage("SRL: " .. msg);
   end
end

local LINEWALKER = {};

local function LINEWALKER_METAIDX(t, i) return LINEWALKER[i]; end

function LINEWALKER:New()
   return self:Bless({});
end

function LINEWALKER:Bless(object, dupeMetatable) 
   local mt = getmetatable(object);
   if (mt) then
      if (copyMetatable) then
	 local nmt = {};
	 for k,v in pairs(mt) do
	    nmt[k] = v;
	 end
	 mt = nmt;
	 setmetatable(object,mt);
      end
   else
      mt = {};
      setmetatable(object, mt);
   end
   local idx = mt.__index;
   if (not idx) then
      mt.__index = LINEWALKER_METAIDX;
   else
      mt.__index = function(t,i) return LINEWALKER[i] or idx(t,i) end
   end

   object.x = object.x or 0;
   object.y = object.y or 0;
   object.moveAngle = object.moveAngle or 0;
   object.angle = object.angle or 0;
   return object;
end

function LINEWALKER:SetLocation(x,y,a,ma) 
   self.x = x or 0;
   self.y = y or 0;
   self.angle = a or 0;
   self.moveAngle = ma or self.angle;
end

function LINEWALKER:GetLocation() 
   return self.x, self.y, self.angle, self.moveAngle;
end

function LINEWALKER:SetContext(context)
   self.context = context;
end

-- The actual walking functions

function LINEWALKER:Turn(angle, stickyFlag)
   self.moveAngle = self.moveAngle + angle;
   if (stickyFlag) then
      self.angle = self.angle + angle;
   end
   --DEBUG("  <W> Turned " .. angle .. ": moveAngle=" .. self.moveAngle);
   --if (stickyFlag) then
   --      DEBUG("  <W>      Sticky turn: angle=" .. self.angle);
   --end
end

function LINEWALKER:Move(relativeTo,distance)
   local scale = 1.0;
   if (relativeTo ~= "R") then
      local relObj = self.context and self.context[relativeTo];
      if (not relObj) then
	 DEBUG("Attempt to move relative to unknown reference '"
	       .. relativeTo .. "'");
	 return;
      end
      scale = relObj.rootScale or 1.0;
   end
   scale = scale * distance;
   local math = math;
   local rad = math.rad(self.moveAngle);
   self.x = self.x + math.sin(rad) * scale; -- RIGHT is positive
   self.y = self.y + math.cos(rad) * scale; -- UP is positive
   --DEBUG("  <W> Moved " .. distance .. " relative to " .. relativeTo
      --.. ": x=" .. self.x .. " y=" .. self.y);
end

-- Instruction list versions

-- turn <angle> [stickyFlag]
function LINEWALKER:turn(ops)
   return self:Turn(ops[2], ops[3]);
end

-- move <relativeTo> <distance>
function LINEWALKER:move(ops)
   return self:Move(ops[2], ops[3]);
end

-- Execution function
function LINEWALKER:Execute(instructionList)
   for _, instruction in ipairs(instructionList) do
      local iType = type(instruction);
      if (iType == "function") then
	 instruction(self);
      elseif (iType == "table") then
	 local op = instruction[1];
	 local opFunc = self[op];
	 if (not opFunc) then
	    DEBUG("Unknown operation '" .. op .. "' in instruction list!");
	 else
	    -- 
	    opFunc(self, instruction);
	 end
      else
	 DEBUG("Found object of type " .. iType .. " in instruction list!");
      end
   end
end

function StatRingsLayout_CreateWalker()
   return LINEWALKER:New();
end

local LAYOUT = {};

local function LAYOUT_METAIDX(t, i) return LAYOUT[i]; end

function LAYOUT:New()
   return self:Bless({});
end

function LAYOUT:Bless(object, dupeMetatable) 
   local mt = getmetatable(object);
   if (mt) then
      if (copyMetatable) then
	 local nmt = {};
	 for k,v in pairs(mt) do
	    nmt[k] = v;
	 end
	 mt = nmt;
	 setmetatable(object,mt);
      end
   else
      mt = {};
      setmetatable(object, mt);
   end
   local idx = mt.__index;
   if (not idx) then
      mt.__index = LAYOUT_METAIDX;
   else
      mt.__index = function(t,i) return LAYOUT[i] or idx(t,i) end
   end

   object.origin = {x=0,y=0,angle=0,rootScale=1.0};
   object.anchors = {};
   object.walker = StatRingsLayout_CreateWalker();
   object.walkContext = {};
   object.walker:SetContext(object.walkContext);

   return object;
end

local function LAYOUT_CDL_FollowChain(objects, entry, workTable)
   local level = workTable[entry];
   if (level) then return level; end
   
   -- Find the parent level
   local entryObject = objects[entry];
   if (not entryObject) then
      level = "missing " .. entry;
   else
      local parent = entryObject.parent;
      if (not parent) then
	 level = "unrooted " .. entry;
      else
	 workTable[entry] = "cycle " .. entry;
	 level = LAYOUT_CDL_FollowChain(objects, parent, workTable);
	 if (type(level) == "number") then
	    level = level + 1;
	 end
      end
   end
   workTable[entry] = level;
   return level;
end

function LAYOUT:CreateDependencyList(layout, workTable)
   if (not workTable) then
      workTable = {};
   end
   workTable[layout.root] = 0;

   local objects = layout.objects;
   local firstErr = nil;

   for entry in pairs(objects) do
      local eLevel = LAYOUT_CDL_FollowChain(objects, entry, workTable);
      if (type(eLevel) ~= "number") then
	 firstErr = eLevel;
      end
   end

   return (firstErr == nil), workTable, firstErr;
end

function LAYOUT:CalculateAnchor(anchorName)
   --DEBUG("Determining anchor location for " .. anchorName);
   local layout = self.layout;
   local objects = layout.objects;
   local objName = anchorName;
   local obj = objects[objName];
   if (not obj) then
      return "Object '" .. objName .. "' not found";
   end
   
   local parAnchor;
   local parName = obj.parent;
   if (parName) then
      parAnchor = self:GetAnchor(parName);
   else
      if (objName == layout.root) then
	 parAnchor = self.origin;
      else
	 return "Object '" .. objName .. "' has no parent";
      end
   end
	 
   local objAnchor = { 
      rootScale = (obj.scale or 1) * parAnchor.rootScale,
   }
   local walkContext = self.walkContext;
   local walker      = self.walker;
   walkContext.O = parAnchor;
   walkContext.S = objAnchor;
   walker:SetLocation(parAnchor.x, 
		      parAnchor.y,
		      parAnchor.angle);
   
   local instructions = obj.anchorPath;
   if (instructions) then
      if (type(instructions) == "function") then
	 instructions(walker);
      else
	 walker:Execute(instructions);
      end
   end
	    
   local x,y,a = walker:GetLocation();
   objAnchor.x = x;
   objAnchor.y = y;
   objAnchor.angle = a;
   return objAnchor;
end

function LAYOUT:GetAnchor(anchorName)
   local anchors = self.anchors;
   local anchor = anchors[anchorName];
   if (not anchor) then
      anchors[anchorName] = "Loop detected with '" .. anchorName .. "'";
      anchor = self:CalculateAnchor(anchorName);
      anchors[anchorName] = anchor;

--[[      if (type(anchor) == "string") then
	 DEBUG("  !!! Anchor '" .. anchorName .. "' failed: "
	       .. anchor);
      else
	 DEBUG("  >>> Anchor for " .. anchorName .. " is ("
	       .. anchor.x .. "," .. anchor.y
		  .. ") - Angle " .. anchor.angle);
      end ]]
   end
   return anchor;
end

function LAYOUT:ProcessLayout(layout)
   self.layout = layout;

--   DEBUG("Compiling layout...");
   self:GetAnchor(layout.root);

   local objects = layout.objects;
   for objName in pairs(objects) do
      self:GetAnchor(objName);
   end
end

function StatRingsLayout_CreateLayoutManager()
   return LAYOUT:New();
end

